/* 
 * Author: Cezary Gwizdowski - T-TIZ
 * biblioteka stl, opis algorytmu : sort
 * File: stl_sort.cpp
 */

/*
 * Algorytm sort umożliwia nam wydajne sortowanie dowolnych zbiorów elementów.
 * Są dwie wersje algorytmu. Pierwsza pobiera dwa argumenty - sort(first,last), 
 * które odpowiednio oznaczają pierwszy i ostatni element pomiędzy
 * którymi mają one zostać posortowane. Element first znajduje się z w zbiorze 
 * elementów sortowanych, a element last już nie - [first,last). Do porównywania 
 * elementów używany jest operator "<" , którego działanie możemy dowolnie 
 * przystosować dla obiektów sortowanych. Druga wersja algorytmu pobiera 
 * dodatkowo trzeci parametr - sort(first,last,comp). Ostatni parametr oznacza
 * funkcję, która zostanie wykorzystana do porównywanie elementów podczas
 * sortowania. Argumentami tej funkcji są dwa elementy ze zbioru. Zwraca ona 
 * true jeśli pierwszy element ma się znaleźć przed drugim i false w przeciwnym 
 * wypadku. Jeśli w sortowanym zbiorze będa elementy równe to nie ma pewności 
 * na to że zostaną one w oryginalnej kolejności.  
 */

#include <stdlib.h>
#include <stdio.h>
#include <iostream>
#include <string>
#include <vector>
#include <sstream>

/* 
 * Algorytm sort biblioteki STL znjaduje się w komponencie algorytm dlatego jeśli 
 * chcemy używać tego algorytmu należy dołączyć plik nagłówkowy <algorytm>. 
 */
#include <algorithm>
/*
 * Biblioteka standardowa (STL) została zawarta w przestrzeni nazw. O tym, że
 * chcemy jej używać informujemy kompilator w następujący sposób:
 */
using namespace std;

/*
 * Klasa odzwierciedla samochód przechowując pewne jego parametry.
 */
class Samochod
{
public:
    /* Nazwa samochodu. */
    string nazwa;   
    /* Moc silnika samochodu. */
    int moc ;
    /* Cena samochodu. */
    int cena;

    /* Konstruktor klasy Samochód. */
    Samochod(string nazwa, int moc, int cena) {
        this->nazwa = nazwa;
        this->moc = moc;
        this->cena = cena;
    }
};

/*
 * Funkcja porównuje moc dwóch obiektów typu Samochod. Gdy moc pierwszego jest
 * większa funkcja zwraca true. W przeciwnym razie zwraca false. 
 */
bool porownajMocSamochodow(Samochod s1, Samochod s2) { 
    if (s1.moc < s2.moc) return true;
    else return false; 
}

/*
 * Klasa klasyfikator, która porównuje samochody pod względem ceny.
 */
struct KlasyfikatorCenowy {
  bool operator() (Samochod s1, Samochod s2) { 
      if (s1.cena < s2.cena) return true;
      else return false;
  }
} mojKlasyfikatorCenowy; //obiekt klasyfikatora cenowego

int main(int argc, char** argv) {
   
    printf("Przykłady użycia algorytmu SORT z biblioteki STL.\n\n");
    
    /*
     * Przykład użycia algorytmu sort z dwoma parametrami na talicy typu int.
     * Do porównywania elementów używany jest domyślny operator "<" dla int.
     */
    printf("1. Tablica typu int:\n\t");
    int a[8] = {5, 29, 1, -40, -17, 65, 100, -100};
    for (int i=0; i<8; i++) {
        printf("%4d ", a[i]);
    }
    /*
     * Out: 5   29    1  -40  -17   65  100 -100
     */
    printf("\n\nPo użyciu algorytmu sort:\n\t");
    sort(a, a+8); //sortowanie
    for (int i=0; i<8; i++) {
        printf("%4d ", a[i]);
    }
    /* 
     * Out: -100  -40  -17    1    5   29   65  100 
     */

    /*
     * Przykład użycia algorytmu sort z dwoma parametrami na talicy typu int.
     * Sortowane są tylko elementy od 2 do 7.
     * Do porównywania elementów używany jest domyślny operator "<" dla int.
     */
    printf("\n\n2. Sortowanie ze zmiana zakresu (2-7). Tablica typu int:\n\t");
    int b[8] = {5, 29, 1, -40, -17, 65, 100, -100};
    for (int i=0; i<8; i++) {
        printf("%4d ", b[i]);
    }
    /*
     * Out: 5   29    1  -40  -17   65  100 -100
     */
    printf("\n\nPo użyciu algorytmu sort:\n\t");
    sort(b+2, b+7); //sortowanie
    for (int i=0; i<8; i++) {
        printf("%4d ", b[i]);
    }
    /* 
     * Out: 5   29  -40  -17    1   65  100 -100 
     *  Witać na tym przykładzie, że zostały posortowane jedynie elementy [2,7). 
     */
    
    /*
     * Przykład użycia algorytmu sort z dwoma parametrami na obiekcie typu 
     * vector przechowującego obiekty typu string. Do porównywania elementów 
     * używany jest domyślny operator "<" dla string.
     */
    printf("\n\n3. Vector przechowujący obiekty typu string:\n");
    vector <string> v1;
    v1.push_back("posiedzieć");
    v1.push_back("Dobrze");
    v1.push_back("Żubrze :)");
    v1.push_back("jest");
    v1.push_back("przy");
    for(int i = 0;i < v1.size();i++){
        cout << "\t" << i << ":" << v1.at(i) << "\n";
    }
    /*
     *	Out: 0:posiedzieć
     *       1:Dobrze
     *       2:Żubrze :)
     *       3:jest
     *       4:przy
     */
    sort(v1.begin(), v1.end()); //sortowanie
    printf("\nPo użyciu algorytmu sort:\n");
    for(int i = 0;i < v1.size();i++){
        cout << "\t" << i << ":" << v1.at(i) << "\n";
    }
    /* 
     * Out: 0:Dobrze
     *      1:jest
     *      2:posiedzieć
     *      3:przy
     *      4:Żubrze :) 
     * W tym wypadku elementy zostały posortowane alfabetycznie.
     */
    
    /*
     * Przykład użycia algorytmu sort na obiektach typu Samochod. 
     * Do porównania tych obiektów zostanie wykorzystana funkcja 
     * porownajMocSamochodow() porównująca moc dwóch samochodów.
     */
    printf("\n4. Vector przechowujący obiekty typu Samochod:\n");
    vector <Samochod> v2;
    v2.push_back(Samochod("Lamborghini Gallardo", 512, 723450));
    v2.push_back(Samochod("Porsche 911", 190, 346000));
    v2.push_back(Samochod("Ferrari 250 S", 210, 1300000));
    for(int i = 0;i < v2.size();i++){
        cout << "\t[Samochód] Nazwa: " << v2.at(i).nazwa << " , Cena: " 
            << v2.at(i).cena << " , Moc: " << v2.at(i).moc << "\n"; 
    }
    sort(v2.begin(), v2.end(), porownajMocSamochodow); //sortowanie
    printf("\nPo użyciu algorytmu sort - porównanie za pomocą funkcji porównującej moc samochodów:\n");
    for(int i = 0;i < v2.size();i++){
        cout << "\t[Samochód] Nazwa: " << v2.at(i).nazwa << " , Cena: " 
            << v2.at(i).cena << " , Moc: " << v2.at(i).moc << "\n"; 
    }
    /* 
     * Out: [Samochód] Nazwa: Porsche 911 , Cena: 346000 , Moc: 190
     *      [Samochód] Nazwa: Ferrari 250 S , Cena: 1300000 , Moc: 210
     *      [Samochód] Nazwa: Lamborghini Gallardo , Cena: 723450 , Moc: 512
     */
    
    /*
     * Przykad użycia obiektu do porównywania elementów podczas sortowania.
     */
    sort(v2.begin(), v2.end(), mojKlasyfikatorCenowy); //sortowanie
    printf("\nPo użyciu algorytmu sort - porównanie za pomocą obiektu klasyfikatora cenowego porównującego cene samochodów:\n");
    for(int i = 0;i < v2.size();i++){
        cout << "\t[Samochód] Nazwa: " << v2.at(i).nazwa << " , Cena: " 
            << v2.at(i).cena << " , Moc: " << v2.at(i).moc << "\n"; 
    }
    /*
     * Out: [Samochód] Nazwa: Porsche 911 , Cena: 346000 , Moc: 190
     *      [Samochód] Nazwa: Lamborghini Gallardo , Cena: 723450 , Moc: 512
     *      [Samochód] Nazwa: Ferrari 250 S , Cena: 1300000 , Moc: 210
     */
    
    return (EXIT_SUCCESS);
}
